#ifndef BUTTON_BOARD_LIB_H
#define BUTTON_BOARD_LIB_H

/* Error Codes */
// General
#define BB_ERR_OK                     0
#define BB_ERR_FAIL                   -1
// SPI
#define BB_ERR_I2C_READ_FAIL          -50
#define BB_ERR_I2C_WRITE_FAIL         -51

#define BB_BTN_COUNT                  5
#define BB_EVENTS_COUNT               10
#define BB_TIMEOUT_MS                 25

typedef signed char err;

typedef struct
{
  uint8_t event;
  uint8_t presses;
} btn_event_t;

typedef btn_event_t btn_events_t[BB_EVENTS_COUNT];

class ButtonBoard
{
  public:
    typedef enum: uint8_t
    {
      BTN_SWITCH_MODE_DISABLED,
      BTN_SWITCH_MODE_CAPTURE,        // Copies input directly
      BTN_SWITCH_MODE_PRESSED,        // Pressed any time
      BTN_SWITCH_MODE_RELEASED,       // Released any time
      BTN_SWITCH_MODE_SHORT_RELEASED, // Released before multi timeout
      BTN_SWITCH_MODE_MULTI_PRESSED,  // Held for multi timeout
      BTN_SWITCH_MODE_MULTI_RELEASED, // Released for multi timeout
      BTN_SWITCH_MODE_HOLD,           // Held for hold timeout
      BTN_SWITCH_MODE_HOLD_RELEASED,  // Released holding for hold timeout
    } E_BTN_SWITCH_MODE;

    typedef enum: uint8_t
    {
      BTN_SWITCH_STATE_SET_UNCHANGED,
      BTN_SWITCH_STATE_SET_ON,
      BTN_SWITCH_STATE_SET_OFF,
      BTN_SWITCH_STATE_SET_CAPTURE,
    } E_BTN_SWITCH_STATE_SET;

    typedef enum: uint8_t
    {
      BTN_MULTI_MODE_THRESHOLD,   // Presses must match or exceed multi_count
      BTN_MULTI_MODE_MATCH        // Presses must match exactly multi_count
    } E_BTN_SWITCH_MULTI_MODE;

    typedef enum: uint8_t
    {
      PORTS_INPUT_INVERTED  = 0x08,
      PORTS_INPUT_ON        = 0x11,
      PORTS_INPUT_OFF       = 0x10
    } E_PORTS_INPUT;
    
    typedef enum: uint8_t
    {
      PORTS_OUTPUT_CH0          = 0x00,
      PORTS_OUTPUT_CH1          = 0x01,
      PORTS_OUTPUT_CH2          = 0x02,
      PORTS_OUTPUT_CH3          = 0x03,
      PORTS_OUTPUT_CH4          = 0x04,
      PORTS_OUTPUT_CH0_INVERTED = 0x08,
      PORTS_OUTPUT_CH1_INVERTED = 0x09,
      PORTS_OUTPUT_CH2_INVERTED = 0x0A,
      PORTS_OUTPUT_CH3_INVERTED = 0x0B,
      PORTS_OUTPUT_CH4_INVERTED = 0x0C,
      PORTS_OUTPUT_ON           = 0x11,
      PORTS_OUTPUT_OFF          = 0x10
    } E_PORTS_OUTPUT;
    
    typedef enum: uint8_t
    {
      LED_OUTPUT_CH0          = 0x00,
      LED_OUTPUT_CH1          = 0x01,
      LED_OUTPUT_CH2          = 0x02,
      LED_OUTPUT_CH3          = 0x03,
      LED_OUTPUT_CH4          = 0x04,
      LED_OUTPUT_CH0_INVERTED = 0x08,
      LED_OUTPUT_CH1_INVERTED = 0x09,
      LED_OUTPUT_CH2_INVERTED = 0x0A,
      LED_OUTPUT_CH3_INVERTED = 0x0B,
      LED_OUTPUT_CH4_INVERTED = 0x0C,
      LED_OUTPUT_ON           = 0x11,
      LED_OUTPUT_OFF          = 0x10
    } E_LED_OUTPUT;
    
    typedef enum: uint8_t
    {
      PIN_PULLS_TRISTATE          = 0x00,
      PIN_PULLS_WEAK_PULL_UP      = 0x01,
      PIN_PULLS_STRONG_PULL_UP    = 0x02,
      PIN_PULLS_STRONG_PULL_DOWN  = 0x03,
    } E_PIN_PULLS;
    
    typedef enum: uint8_t
    {
      CMD_IDLE                  = 0x00,
      CMD_APPLY_CONFIG          = 0x01,
      CMD_SAVE_CONFIG           = 0x02,
      CMD_LOAD_CONFIG           = 0x03,
      CMD_LOAD_DEFAULT_CONFIG   = 0x04,
      CMD_RESET                 = 0x05,
    } E_CMD;
    
    typedef enum: uint8_t
    {
      UART_CONFIG_DISABLED      = 0x00,
      UART_CONFIG_MASK_TEXT     = 0x01,
      UART_CONFIG_MASK_HEX      = 0x02,
    } E_UART_CONFIG;
    
    typedef struct __attribute__((packed))
    {
      uint8_t pin_inputs;
      uint8_t btn_inputs;
      uint8_t switch_outputs;
      uint8_t pin_outputs;
    } port_values_t;
    
    typedef struct __attribute__((packed))
    {
      uint8_t addr_link_open;
      uint8_t addr_link_short;
      uint8_t addr_override;
    } addr_t;
    
    ButtonBoard( uint8_t pin_ready_init, uint8_t i2c_addr_init );
    
    bool Ready( void );
    err ReadVersion( uint8_t *version );
    err WriteSwitchMode( E_BTN_SWITCH_MODE *switch_mode, uint8_t num );
    err WriteSwitchMultiMode( E_BTN_SWITCH_MULTI_MODE *switch_multi_mode, uint8_t num );
    err WriteSwitchStateSet( E_BTN_SWITCH_STATE_SET *switch_state_set, uint8_t num );
    err WriteSwitchMultiCount( uint8_t *multi_count, uint8_t num );
    err WritePortsInput( E_PORTS_INPUT *ports_input, uint8_t num );
    err WritePortsOutput( E_PORTS_OUTPUT *ports_output, uint8_t num );
    err WritePinPulls( E_PIN_PULLS *pin_pulls, uint8_t num );
    err WriteLedOutput( E_LED_OUTPUT led_output );
    err ReadEvents( btn_events_t *events );
    err WriteDebounceMs( uint8_t *debounce_ms, uint8_t num );
    err WriteMultiTimeoutMs( uint16_t *multi_timeout_ms, uint8_t num );
    err WriteHoldTimeoutMs( uint16_t *hold_timeout_ms, uint8_t num );
    err WriteCmd( E_CMD cmd );
    err ReadPorts( port_values_t *port_values );
    err WriteUartConfig( E_UART_CONFIG uart_config );
    err WriteAddr( addr_t *addr );
    err ReadAddrActive( uint8_t *addr );
    
  private:
    typedef enum: uint8_t
    {
      I2C_REGS_VERSION,               // 00
      I2C_REGS_SWITCH_MODE,           // 01
      I2C_REGS_SWITCH_MULTI_MODE,     // 02
      I2C_REGS_SWITCH_MULTI_COUNT,    // 03
      I2C_REGS_SWITCH_STATE_SET,      // 04
      I2C_REGS_PORTS_INPUT,           // 05
      I2C_REGS_PORTS_OUTPUT,          // 06
      I2C_REGS_PIN_PULLS,             // 07
      I2C_REGS_LED_OUTPUT,            // 08
      I2C_REGS_EVENTS,                // 09
      I2C_REGS_DEBOUNCE_MS,           // 10
      I2C_REGS_MULTI_TIMEOUT_MS,      // 11
      I2C_REGS_HOLD_TIMEOUT_MS,       // 12
      I2C_REGS_CMD,                   // 13
      I2C_REGS_PORTS_READ,            // 14
      I2C_REGS_UART_CONFIG,           // 15
      I2C_REGS_I2C_ADDR,              // 16
      I2C_REGS_I2C_ADDR_ACTIVE,       // 17
    } E_I2C_REGS;
    
    uint8_t pin_ready;
    uint8_t i2c_addr;
    
    err i2c_read_reg( uint8_t reg_num, uint8_t num_bytes, uint8_t *data );
    err i2c_write_reg( uint8_t reg_num, uint8_t num_bytes, uint8_t *data );
};

#endif