#include <Arduino.h>
#include <Wire.h>
#include "ButtonBoard.h"

void clip_byte( uint8_t *value, uint8_t max )
{
  if ( *value > max )
    *value = max;
}

ButtonBoard::ButtonBoard( uint8_t pin_ready_init, uint8_t i2c_addr_init )
{
  pin_ready = pin_ready_init;
  i2c_addr = i2c_addr_init;
  
  pinMode( pin_ready, INPUT );
  
  Wire.setWireTimeout( (uint16_t)1000 * BB_TIMEOUT_MS /* us */, true /* reset_on_timeout */ );
}

bool ButtonBoard::Ready( void )
{
  return ( digitalRead( A3 ) == 0 );
}

err ButtonBoard::ReadVersion( uint8_t *version )
{
  return i2c_read_reg( I2C_REGS_VERSION, 1, version );
}

err ButtonBoard::WriteSwitchMode( E_BTN_SWITCH_MODE *switch_mode, uint8_t num )
{
  clip_byte( &num, BB_BTN_COUNT );
  
  return i2c_write_reg( I2C_REGS_SWITCH_MODE, num, (uint8_t *)switch_mode );
}

err ButtonBoard::WriteSwitchMultiMode( E_BTN_SWITCH_MULTI_MODE *switch_multi_mode, uint8_t num )
{
  clip_byte( &num, BB_BTN_COUNT );
  
  return i2c_write_reg( I2C_REGS_SWITCH_MULTI_MODE, num, (uint8_t *)switch_multi_mode );
}

err ButtonBoard::WriteSwitchStateSet( E_BTN_SWITCH_STATE_SET *switch_state_set, uint8_t num )
{
  clip_byte( &num, BB_BTN_COUNT );
  
  return i2c_write_reg( I2C_REGS_SWITCH_STATE_SET, num, (uint8_t *)switch_state_set );
}

err ButtonBoard::WriteSwitchMultiCount( uint8_t *multi_count, uint8_t num )
{
  clip_byte( &num, BB_BTN_COUNT );
  
  return i2c_write_reg( I2C_REGS_SWITCH_MULTI_COUNT, num, multi_count );
}

err ButtonBoard::WritePortsInput( E_PORTS_INPUT *ports_input, uint8_t num )
{
  clip_byte( &num, BB_BTN_COUNT );
  
  return i2c_write_reg( I2C_REGS_PORTS_INPUT, num, (uint8_t *)ports_input );
}

err ButtonBoard::WritePortsOutput( E_PORTS_OUTPUT *ports_output, uint8_t num )
{
  clip_byte( &num, BB_BTN_COUNT );
  
  return i2c_write_reg( I2C_REGS_PORTS_OUTPUT, num, (uint8_t *)ports_output );
}

err ButtonBoard::WritePinPulls( E_PIN_PULLS *pin_pulls, uint8_t num )
{
  clip_byte( &num, BB_BTN_COUNT );
  
  return i2c_write_reg( I2C_REGS_PIN_PULLS, num, (uint8_t *)pin_pulls );
}

err ButtonBoard::WriteLedOutput( E_LED_OUTPUT led_output )
{
  return i2c_write_reg( I2C_REGS_LED_OUTPUT, 1, (uint8_t *)&led_output );
}

err ButtonBoard::ReadEvents( btn_events_t *events )
{
  return i2c_read_reg( I2C_REGS_EVENTS, sizeof(btn_events_t), (uint8_t *)events );
}

err ButtonBoard::WriteDebounceMs( uint8_t *debounce_ms, uint8_t num )
{
  clip_byte( &num, BB_BTN_COUNT );
  
  return i2c_write_reg( I2C_REGS_DEBOUNCE_MS, num, debounce_ms );
}

err ButtonBoard::WriteMultiTimeoutMs( uint16_t *multi_timeout_ms, uint8_t num )
{
  clip_byte( &num, BB_BTN_COUNT );
  
  return i2c_write_reg( I2C_REGS_MULTI_TIMEOUT_MS, num << 1, (uint8_t *)multi_timeout_ms );
}

err ButtonBoard::WriteHoldTimeoutMs( uint16_t *hold_timeout_ms, uint8_t num )
{
  clip_byte( &num, BB_BTN_COUNT );
  
  return i2c_write_reg( I2C_REGS_HOLD_TIMEOUT_MS, num << 1, (uint8_t *)hold_timeout_ms );
}

err ButtonBoard::WriteCmd( E_CMD cmd )
{
  return i2c_write_reg( I2C_REGS_CMD, 1, (uint8_t *)cmd );
}

err ButtonBoard::ReadPorts( port_values_t *port_values )
{
  return i2c_read_reg( I2C_REGS_PORTS_READ, sizeof(port_values_t), (uint8_t *)port_values );
}

err ButtonBoard::WriteUartConfig( E_UART_CONFIG uart_config )
{
  return i2c_write_reg( I2C_REGS_UART_CONFIG, 1, (uint8_t *)uart_config );
}

err ButtonBoard::WriteAddr( addr_t *addr )
{
  return i2c_write_reg( I2C_REGS_I2C_ADDR, sizeof(addr_t), (uint8_t *)addr );
}

err ButtonBoard::ReadAddrActive( uint8_t *addr )
{
  return i2c_read_reg( I2C_REGS_I2C_ADDR_ACTIVE, 1, addr );
}

/* Private ----------------------------------------------------------------------- */

err ButtonBoard::i2c_read_reg( uint8_t reg_num, uint8_t num_bytes, uint8_t *data )
{
  err rc;
  uint32_t timestamp;
  bool data_available;
  
  rc = BB_ERR_OK;
  
  Wire.clearWireTimeoutFlag();
  
//  Wire.beginTransmission( device_id );
//  Wire.write( reg_num );
//  Wire.endTransmission( false );  // Repeated start

//  Wire.requestFrom( device_id, num_bytes );
  
  Wire.requestFrom( i2c_addr, num_bytes, reg_num, 1, true );
  
  if ( Wire.getWireTimeoutFlag() )
    rc = BB_ERR_I2C_READ_FAIL;
  else
  {
    timestamp = millis();
    
    do
    {
      data_available = ( Wire.available() >= num_bytes );
    }
    while ( ( !data_available ) && ( ( millis() - timestamp ) < BB_TIMEOUT_MS ) );
    
    if ( !data_available )
      rc = BB_ERR_I2C_READ_FAIL;
    else
    {
      for ( uint8_t i=0; i<num_bytes; i++ )
        *data++ = Wire.read();
    }
  }
  
  return rc;
}

err ButtonBoard::i2c_write_reg( uint8_t reg_num, uint8_t num_bytes, uint8_t *data )
{
  err rc;
  uint8_t result;
  uint32_t timestamp;
  bool data_available;
  
  rc = BB_ERR_OK;
  
  Wire.clearWireTimeoutFlag();
  
  Wire.beginTransmission( i2c_addr );
  Wire.write( reg_num );
  Wire.write( data, num_bytes );
  result = Wire.endTransmission( true );
  
  if ( result != 0 )
    rc = BB_ERR_I2C_WRITE_FAIL;
  
  return rc;
}
